class GeneralGameData:
    def __init__(self, game_name="", score=0, date_started=None, wins=0, loses=0):
        self.game_name = game_name
        self.score = score
        self.date_started = date_started
        self.wins = wins
        self.loses = loses



jsonable = (bool, str, int, float, type(None), type([]), type({}))


def is_jsonable(obj):
    """
    checks whether the data can be processed with json or not
    :param obj:
    :return:
    """
    return isinstance(obj, jsonable)


def stringify_game_history_dict(game_history_dict):
    """
    stringifies all the keys in game_history dict if they are not primitive type or array or dict
    :param game_history_dict:
    :return: dictionary
    """
    for key in game_history_dict.keys():
        if not is_jsonable(game_history_dict[key]):
            game_history_dict[key] = str(game_history_dict[key])

    return game_history_dict



from datetime import datetime
import mwt_games_manager.managers as mg
from mwt_games_manager.models.feature import Feature
from mwt_games_manager.models.product import Product


def add_product(product):
    """
    Add a product to the database
    :param product:
    :return:
    """
    mg.client.collection("products").document(product.product_id).add(product.__dict__)


def update_product(product):
    """
    Update a product from the database
    :param product:
    :return:
    """
    mg.client.collection("products").document(product.product_id).add(product.__dict__)


def delete_product(product_id):
    """
    Delete a product from the database
    :param product_id:
    :return:
    """
    mg.client.collection("products").document(product_id).delete()


def get_product(product_id):
    """
    Retrieve a product from the database
    :param product_id:
    :return:
    """
    product = mg.client.collection("products").document(product_id).get()
    product = Product(**product.to_dict())
    return product


def get_all_products():
    """
    Retrieve all the products from the database
    :return:
    """
    products = list(mg.client.collection("products").stream())
    products = [Product(**product.to_dict()) for product in products]
    return products


def add_product_feature(product_id, feature):
    """
    Add a product feature to the product
    :param product_id:
    :param feature:
    :return:
    """
    mg.client.collection("products").document(product_id).collection("features").add(feature.__dict__)


def get_product_features(product_id):
    """
    Retrieve all the product features from the database
    :param product_id:
    :return:
    """
    features = list(mg.client.collection("products").document(product_id).collection("features"))
    features = [Feature(**feature) for feature in features]
    return features


def add_user_product(username, product):
    """
    Add a product under a username
    :param username:
    :param product:
    :return:
    """
    mg.client.collection("users").document(username).collection("products").document(product.product_id).set(product.__dict__)


def update_user_product(username, product):
    """
    Update a product under a username
    :param username:
    :param product:
    :return:
    """
    mg.client.collection("users").document(username).collection("products").document(product.product_id).set(product.__dict__)


def delete_user_product(username, product_id):
    """
    Delete a product under a username
    :param username:
    :param product_id:
    :return:
    """
    mg.client.collection("users").document(username).collection("products").document(product_id).delete()


def get_user_product(username, product_id):
    """
    Get a product under a username
    :param username:
    :param product_id:
    :return:
    """
    product = mg.client.collection("users").document(username).collection("products").document(product_id).get()
    product = Product(**product.to_dict())
    return product



from flask import Blueprint, Response, request, jsonify
import mwt_games_manager.managers as managers
import mwt_games_manager.utils as utils

mwt_game_manager_blueprint = Blueprint('mwt-game-manager', __name__)


@mwt_game_manager_blueprint.route('/get-all-user-history', methods=['GET'])
def get_all_user_history():
    if not managers.client:
        raise Exception("module has not been setup properly")

    username = request.args.get("username")
    if not username:
        return Response(response="No username provided", status=400)

    game_histories = list(managers.client.collection("users").document(username).collection("game-data").document(
        managers.default_game_name).collection("history").stream())
    game_histories = [utils.stringify_game_history_dict(game_history.to_dict()) for game_history in game_histories]
    return jsonify(game_histories)


@mwt_game_manager_blueprint.route('/get-user-data', methods=['GET'])
def get_user_data():
    if not managers.client:
        raise Exception("module has not been setup properly")

    username = request.args.get("username")
    if not username:
        return Response(response="No username provided", status=400)

    game_data = managers.client.collection("users").document(username).collection("game-data").document(managers.default_game_name).get()
    return jsonify(game_data.to_dict())



import datetime

import firebase_admin
from firebase_admin import firestore, get_app
from google.cloud.firestore_v1 import FieldFilter

from mwt_games_manager.models.game_history import GameHistory
from mwt_games_manager.models.general_game_data import GeneralGameData
from mwt_games_manager.models.user import User
from mwt_games_manager.models.feature import Feature
from mwt_games_manager.models.product import Product
from flask_bcrypt import check_password_hash

client = None
try:
    client = firestore.client(get_app("games-manager"))
except Exception as e:
    pass

default_game_name = ""


def setup_module(database_credentials, game_name):
    """
    required for initializing the sdk and setting up a default game name
    :param database_credentials:
    :param game_name:
    :return:
    """
    global client, default_game_name
    firestore_app = firebase_admin.initialize_app(database_credentials, name="games-manager")
    client = firestore.client(firestore_app)
    default_game_name = game_name


def is_user_first_game(username, game_name=False):
    """
    checks whether the user has played this game before or not
    :param username:
    :param game_name:
    :return:
    """
    global client, default_game_name
    if not game_name:
        game_name = default_game_name
    if not client:
        raise Exception("module has not been setup properly")

    game_data = client.collection("users").document(username).collection("game-data").document(game_name).get().exists
    return not game_data


def setup_user_game_data(username, game_data=GeneralGameData(), game_name=False):
    """
    setting up game data for a user
    :param username:
    :param game_data:
    :param game_name:
    :return:
    """
    global client, default_game_name
    if not game_name:
        game_name = default_game_name
    if not client:
        raise Exception("module has not been setup properly")

    if not game_data.game_name:
        game_data.game_name = game_name
    if not game_data.date_started:
        game_data.date_started = datetime.datetime.now()
    client.collection("users").document(username).collection("game-data").document(game_name).set(game_data.__dict__)
    client.collection(game_name).document(username).set(game_data.__dict__)


def update_user_game_data(username, game_data, game_name=False):
    """
    updating the game data for a user
    :param username:
    :param game_data:
    :param game_name:
    :return:
    """
    global client, default_game_name
    if not game_name:
        game_name = default_game_name
    if not client:
        raise Exception("module has not been setup properly")

    client.collection("users").document(username).collection("game-data").document(game_name).set(game_data.__dict__)


def get_user_game_data(username, game_name=False):
    """
    getting the game data for a user
    :param username:
    :param game_name:
    :return:
    """
    global client, default_game_name
    if not game_name:
        game_name = default_game_name
    if not client:
        raise Exception("module has not been setup properly")

    game_data = client.collection(game_name).document(username).get()
    if not game_data.exists:
        game_data = get_user_game_data_archive(username, game_name)
        if game_data:
            client.collection(game_name).document(username).set(game_data.__dict__)
        if not game_data:
            return False
    return GeneralGameData(**game_data.to_dict())


def get_user_game_data_archive(username, game_name=False):
    """
    getting the game data for a user
    :param username:
    :param game_name:
    :return:
    """
    global client, default_game_name
    if not game_name:
        game_name = default_game_name
    if not client:
        raise Exception("module has not been setup properly")

    game_data = client.collection("users").document(username).collection("game-data").document(game_name).get()
    if not game_data.exists:
        return False
    return GeneralGameData(**game_data.to_dict())


def _initial_check(username, game_name=False):
    global client, default_game_name
    if not game_name:
        game_name = default_game_name
    if not client:
        raise Exception("module has not been setup properly")

    if is_user_first_game(username, game_name=game_name):
        setup_user_game_data(username, game_name=game_name)


def add_game_history(username, game_history, game_name=False):
    """
    adding a game history for the user
    :param username:
    :param game_history:
    :param game_name:
    :return:
    """
    global default_game_name
    if not game_name:
        game_name = default_game_name
    _initial_check(username, game_name)
    client.collection("users").document(username).collection("game-data").document(game_name).collection(
        "history").document(game_history.game_id).set(game_history.__dict__)


def update_game_history(username, game_history, game_name=False):
    """
    updating a game history for the user
    :param username:
    :param game_history:
    :param game_name:
    :return:
    """
    global default_game_name
    if not game_name:
        game_name = default_game_name
    _initial_check(username, game_name)
    client.collection("users").document(username).collection("game-data").document(game_name).collection(
        "history").document(game_history.game_id).set(game_history.__dict__)


def add_user_win_game(username, game_id, game_name=False, new_status="win", added_points=10):
    """
    This will automatically update the game history and add points to the user's game data
    game history: Set status of the game and end time
    game data: add to score and add to the wins
    :param username:
    :param game_id:
    :param game_name:
    :param new_status:
    :return:
    """

    global default_game_name
    if not game_name:
        game_name = default_game_name

    game_history = get_game_history(username, game_id, game_name)
    game_history.status = new_status
    game_history.finished_date = datetime.datetime.now()
    update_game_history(username, game_history, game_name)

    game_data = get_user_game_data(username, game_name)
    game_data.wins += 1
    game_data.score += added_points
    update_user_game_data(username, game_data, game_name)


def add_user_lose_game(username, game_id, game_name=False, new_status="lose", deducted_points=10):
    """
    This will automatically update the game history and add points to the user's game data
    game history: Set status of the game and end time
    game data: add to score and add to the wins
    :param username:
    :param game_id:
    :param game_name:
    :param new_status:
    :param deducted_points:
    :return:
    """

    global default_game_name
    if not game_name:
        game_name = default_game_name

    game_history = get_game_history(username, game_id, game_name)
    game_history.status = new_status
    game_history.finished_date = datetime.datetime.now()
    update_game_history(username, game_history, game_name)

    game_data = get_user_game_data(username, game_name)
    game_data.loses -= 1
    game_data.score -= deducted_points
    update_user_game_data(username, game_data, game_name)


def get_all_users_by_score(limit=False, game_name=False):
    """
    Will return in list of users in descending score order
    :param limit:
    :param game_name:
    :return:
    """

    global default_game_name
    if not game_name:
        game_name = default_game_name

    client.collection(game_name).order_by("score", direction=firestore.Query.DESCENDING)


def delete_game_history(username, game_id, game_name=False):
    """
    deleting a game history for the user
    :param username:
    :param game_id:
    :param game_name:
    :return:
    """
    global default_game_name
    if not game_name:
        game_name = default_game_name
    _initial_check(username, game_name)
    client.collection("users").document(username).collection("game-data").document(game_name).collection(
        "history").document(game_id).delete()


def get_game_history(username, game_id, game_name=False):
    """
    getting a game history for the user
    :param username:
    :param game_id:
    :param game_name:
    :return:
    """
    global default_game_name
    if not game_name:
        game_name = default_game_name
    _initial_check(username, game_name)
    game_history = client.collection("users").document(username).collection("game-data").document(game_name).collection(
        "history").document(game_id).get().to_dict()

    if game_history is None:
        return False

    game_history = GameHistory(**game_history)
    return game_history


def validate_user(username, password, game_name=False):
    """
    authenticate whether username and password are valid
    :param username:
    :param password:
    :param game_name:
    :return:
    """
    global default_game_name
    if not game_name:
        game_name = default_game_name
    _initial_check(username, game_name)
    user = client.collection("users").document(username).get()
    if not user.exists:
        return False

    user = User(**user.to_dict())

    return check_password_hash(user.password, password)


def fetch_user(username):
    """
    fetching the user information for the user
    :param username:
    :return:
    """
    doc = client.collection("users").document(username).get()
    if not doc.exists:
        return False

    user = User(**doc.to_dict())
    return user


def get_all_game_data(username):
    """
    fetching game data for all the games
    :param username:
    :return:
    """
    docs = list(client.collection("users").document(username).collection("game-data").stream())
    game_datas = [GeneralGameData(**game_data.to_dict()) for game_data in docs]
    return game_datas


def get_features(username, game_name=None):
    """
    fetches all the features the user has for a specific game
    :param username:
    :param game_name:
    :return:
    """
    global default_game_name
    if not game_name:
        game_name = default_game_name
    products = list(client.collection("users").document(username).collection("products").stream())
    products = [Product(**product.to_dict()) for product in products]

    features = []

    for product in products:
        product_features = list(client.collection("products").document(product.product_id).collection("features").where(
            filter=FieldFilter("game", "==", game_name)).stream())
        product_features = [Feature(**feature.to_dict()) for feature in product_features]
        features.append(product_features)

    return features
